/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>
#include "palmpalette.h"
#include "pgpf.h"

//#define DEBUG

typedef struct _Palette Palette;
struct _Palette {
	int id, ncolors;
	guint32 *color;
};

GHashTable *images, *frames, *tiles, *palettes;
char *comment = NULL;
const Palette default_palette = {
	0, sizeof(palm_palette)/sizeof(palm_palette[0]),
	palm_palette
};

static void usage(const char *prog) {
	fprintf(stderr, "Usage:\n");
	fprintf(stderr, "  %s [options] foo.prc > foo.ppm\n\n", prog);
	fprintf(stderr, "Options:\n");
	fprintf(stderr, "  -t     Decode thumbnail instead of main image\n");
	exit(1);
}
static void pgpf2ppm(FILE *ifp, int imgid);
static void read_inf(FILE *ifp, int id);
static void read_fri(FILE *ifp, int id);
static void read_til(FILE *ifp, int id);
static void read_dat(FILE *ifp, int id);
static void read_plt(FILE *ifp, int id);
static void decode_image(FILE *ifp, Image *img);
static void decode_tile(Tile *tile, FILE *ifp, unsigned char *buf,
		const Palette *palette, int bpp, int compr);

int main(int argc, char **argv) {
	FILE *fp;
	char c;
	int imgid = IMG_MAIN;

	while ((c = getopt(argc, argv, "t")) != -1)
		switch (c) {
			case 't':
				imgid = IMG_THUMB;
				break;
			default:
				usage(argv[0]);
		}
	if (optind != argc-1)
		usage(argv[0]);

	if ((fp = fopen(argv[optind], "r")) == NULL) {
		perror(argv[optind]);
		exit(1);
	}

	images = g_hash_table_new(g_direct_hash, g_direct_equal);
	frames = g_hash_table_new(g_direct_hash, g_direct_equal);
	tiles = g_hash_table_new(g_direct_hash, g_direct_equal);
	palettes = g_hash_table_new(g_direct_hash, g_direct_equal);

	pgpf2ppm(fp, imgid);
	fclose(fp);

	return 0;
}

static void pgpf2ppm(FILE *ifp, int imgid) {
	char name[33];
	unsigned short attributes, version;
	unsigned int c_date, m_date, b_date, mod_number, appinfo, sortinfo;
	char dbtype[5], dbcreator[5];
	unsigned int uniqueid, nextrecordlist;
	unsigned short numrecords;
	int i;

	fread(name, 1, 32, ifp);  name[32] = '\0';
	attributes = readint(ifp, 2);
	version = readint(ifp, 2);
	c_date = readint(ifp, 4);
	m_date = readint(ifp, 4);
	b_date = readint(ifp, 4);
	mod_number = readint(ifp, 4);
	appinfo = readint(ifp, 4);
	sortinfo = readint(ifp, 4);
	fread(dbtype, 1, 4, ifp);  dbtype[4] = '\0';
	fread(dbcreator, 1, 4, ifp);  dbcreator[4] = '\0';
	uniqueid = readint(ifp, 4);
	nextrecordlist = readint(ifp, 4);
	numrecords = readint(ifp, 2);

#ifdef DEBUG
	fprintf(stderr, "name=\"%s\" attr=%x ver=%x cdate=%d mdate=%d bdate=%d\n",
		name, attributes, version, c_date, m_date, b_date);
	fprintf(stderr, "mod_number=%d appinfo=%x sortinfo=%x\n",
		mod_number, appinfo, sortinfo);
	fprintf(stderr, "type=\"%s\" creator=\"%s\" id=%d nextrec=%d numrec=%d\n",
		dbtype, dbcreator, uniqueid, nextrecordlist, numrecords);
#endif

	for (i=0; i<numrecords; i++) {
		char type[5];
		int pos;
		unsigned short id;
		unsigned int loc;
		fread(type, 1, 4, ifp);  type[4] = '\0';
		id = readint(ifp, 2);
		loc = readint(ifp, 4);
#ifdef DEBUG
		fprintf(stderr, "rec[%d]: type=\"%s\" id=%d loc=%x\n", i, type, id, loc);
#endif
		pos = ftell(ifp);
		fseek(ifp, loc, SEEK_SET);
		if (!strcmp(type, "iINF")) read_inf(ifp, id);
		else if (!strcmp(type, "iFRI")) read_fri(ifp, id);
		else if (!strcmp(type, "iTIL")) read_til(ifp, id);
		else if (!strcmp(type, "iDAT")) read_dat(ifp, id);
		else if (!strcmp(type, "iPLT")) read_plt(ifp, id);
		else fprintf(stderr, "Resource %d type \"%s\" ignored.\n", i, type);
		fseek(ifp, pos, SEEK_SET);
	}

	decode_image(ifp, g_hash_table_lookup(images, GUINT_TO_POINTER(imgid)));
}

static void read_inf(FILE *ifp, int id) {
	unsigned short dummy1, dummy2;
	Image *img = g_new(Image, 1);

	img->id = id;
	img->width = readint(ifp, 2);
	img->height = readint(ifp, 2);
	dummy1 = readint(ifp, 2);
	img->bpp = readint(ifp, 2);
	dummy2 = readint(ifp, 2);
	img->tilesize = readint(ifp, 2);
	img->compr = readint(ifp, 2);
	img->frameid = readint(ifp, 2);
	img->paletteid = readint(ifp, 2);

	g_hash_table_insert(images, GUINT_TO_POINTER(img->id), img);

#ifdef DEBUG
	fprintf(stderr, "Image: %dx%d @%dbpp tile=%dx%d compr=%x id=%d pal=%d\n",
		img->width, img->height, img->bpp, img->tilesize, img->tilesize,
		img->compr, img->frameid, img->paletteid);
#endif
}

static void read_fri(FILE *ifp, int id) {
	unsigned short wait, flags;
	int i;
	Frame *frame = g_new(Frame, 1);

	frame->id = id;
	frame->nframes = readint(ifp, 2);
	frame->tileid = g_new(int, frame->nframes);
	wait = readint(ifp, 2);
	flags = readint(ifp, 2);
#ifdef DEBUG
	fprintf(stderr, "Frame: %d frames, wait %dms, flags=%x\n",
		frame->nframes, wait, flags);
#endif
	for (i=0; i<frame->nframes; i++) {
		frame->tileid[i] = readint(ifp, 2);
#ifdef DEBUG
		fprintf(stderr, "  tilebeginid=%d\n", frame->tileid[i]);
#endif
	}

	g_hash_table_insert(frames, GUINT_TO_POINTER(frame->id), frame);
}

static void read_til(FILE *ifp, int id) {
	Tile *tile = g_new(Tile, 1);

	tile->id = id;
	tile->length = readint(ifp, 2);
	tile->pos = ftell(ifp);

#ifdef DEBUG
	fprintf(stderr, "Tile: length=%d\n", tile->length);
#endif

	g_hash_table_insert(tiles, GUINT_TO_POINTER(tile->id), tile);
}

static void read_dat(FILE *ifp, int id) {
	char date_time[21];

	fread(date_time, 1, 20, ifp);
	date_time[20] = '\0';
#ifdef DEBUG
	fprintf(stderr, "Creation date: \"%s\"\n", date_time);
#endif
	comment = g_strconcat("DATE: ", date_time, NULL);
}

static void read_plt(FILE *ifp, int id) {
	int i;
	Palette *palette = g_new(Palette, 1);
	palette->id = id;
	palette->ncolors = readint(ifp, 2);
	palette->color = g_new(guint32, palette->ncolors);
	for (i=0; i<palette->ncolors; i++)
		palette->color[i] = readint(ifp, 4);
#ifdef DEBUG
	fprintf(stderr, "Palette: %d colors\n", palette->ncolors);
#endif
	g_hash_table_insert(palettes, GUINT_TO_POINTER(palette->id), palette);
}

static void decode_image(FILE *ifp, Image *img) {
	unsigned char *buf = g_new(unsigned char, img->width * img->height * 3);
	Frame *frm = g_hash_table_lookup(frames, GINT_TO_POINTER(img->frameid));
	const Palette *plt;
	int x, y, tileno = frm->tileid[0];
	
	if (img->bpp == 16)
		plt = NULL;
	else if (img->paletteid)
		plt = g_hash_table_lookup(palettes, GINT_TO_POINTER(img->paletteid));
	else
		plt = &default_palette;

	for (y=0; y<img->height; y+=img->tilesize)
		for (x=0; x<img->width; x+=img->tilesize) {
			Tile *tile = g_hash_table_lookup(tiles, GINT_TO_POINTER(tileno));
			unsigned char *tilebuf = g_new(unsigned char, img->tilesize * img->tilesize * 3);
			int sx, sy;
			decode_tile(tile, ifp, tilebuf, plt, img->bpp, img->compr);
			for (sy=0; sy<img->tilesize && y+sy<img->height; sy++)
				for (sx=0; sx<img->tilesize && x+sx<img->width; sx++)
					memcpy(&buf[((y + sy)*img->width + x + sx)*3], &tilebuf[((sy*img->tilesize)+sx)*3], 3);
			g_free(tilebuf);
			tileno++;
		}

	printf("P6\n");
	if (comment) printf("# %s\n", comment);
	printf("%d %d 255\n", img->width, img->height);
	fflush(stdout);
	write(STDOUT_FILENO, buf, img->width*img->height*3);
	g_free(buf);
}

static void pixel_apply(int pix, const Palette *palette, int bpp,
		unsigned char **buf) {
	if (bpp == 8) {
		guint32 color;
		g_assert(pix >= 0 && pix < palette->ncolors);
		color = palette->color[pix];
		*(*buf)++ = (color & 0x00FF0000) >> 16;  /* red */
		*(*buf)++ = (color & 0x0000FF00) >> 8;   /* green */
		*(*buf)++ = (color & 0x000000FF);        /* blue */
	}
	else {
		*(*buf)++ = (pix & 0xF800) >> 11 << 3;  /* red, 5 bits */
		*(*buf)++ = (pix & 0x07E0) >> 5 << 2;   /* green, 6 bits */
		*(*buf)++ = (pix & 0x001F) << 3;        /* blue, 5 bits */
	}
}

static void decode_tile(Tile *tile, FILE *ifp, unsigned char *buf,
		const Palette *palette, int bpp, int compr) {
	int last_pixel = -1, current_pixel, n = 0;

	g_assert((bpp == 16 && palette == NULL) || (bpp == 8 && palette != NULL));
	g_assert(compr == COMPR_NONE || compr == COMPR_RLE);

	fseek(ifp, tile->pos, SEEK_SET);
	while (n<tile->length) {
		current_pixel = readint(ifp, bpp/8);  n+=bpp/8;
		pixel_apply(current_pixel, palette, bpp, &buf);
		if (compr == COMPR_RLE && current_pixel == last_pixel) {
			int nrep = readint(ifp, 1);     n++;
			while (nrep-- > 0)
				pixel_apply(current_pixel, palette, bpp, &buf);
			last_pixel = -1;
		}
		else
			last_pixel = current_pixel;
	}
}
