/*
 * Copyright (C) 2003 Patrick Reynolds <reynolds@cs.duke.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the
 *      Free Software Foundation, Inc.
 *      59 Temple Place
 *      Suite 330
 *      Boston, MA 02111-1307
 *      USA
 */

/** Floyd and Steinberg dithering method added by Franck Sicard <Franck@3demi.net> */

#include <stdlib.h>
#include "libppm.h"
#include "palmpalette.h"

#define CLAMP(x, low, high) (((x) > (high)) ? (high) : (((x) < (low)) ? (low) : (x)))

static void propagateError(struct ppm_image *ppm, int x, int y, int error, int plane);

void ppmdither(struct ppm_image *ppm) {
	unsigned char pal216map[216];
	//unsigned char palgreymap[16] = { 230, 215, 216, 190, 217, 218, 165, 219, 220, 50, 221, 222, 25, 223, 224, 0 };
	int i, x, y;
	unsigned char *buf;

	/* get a mapping from a red-then-green-then-blue palette to the palm
	 * palette */
	for (i=0; i<256; i++) {
		int red =   (palm_palette[i] & 0x00FF0000) >> 16;
		int green = (palm_palette[i] & 0x0000FF00) >> 8;
		int blue =  palm_palette[i] & 0x000000FF;
		if (red % 0x33 != 0 || green % 0x33 != 0 || blue % 0x33 != 0) continue;
		red /= 0x33;
		green /= 0x33;
		blue /= 0x33;
		pal216map[red*36 + green*6 + blue] = i;
	}

	buf = (unsigned char*)malloc(ppm->width * ppm->height);

	for (y=0; y<ppm->height; y++)
		for (x=0; x<ppm->width; x++) {
			int red =   PPM_PIXEL(*ppm, x, y, 0);
			int green = PPM_PIXEL(*ppm, x, y, 1);
			int blue =  PPM_PIXEL(*ppm, x, y, 2);
			int map_red =   red/0x33;
			int map_green = green/0x33;
			int map_blue =  blue/0x33;          
			int pix = map_red*36 + map_green*6 + map_blue;
			buf[y*ppm->width+x] = pal216map[pix];
			propagateError(ppm, x, y, red - map_red*0x33, 0);
			propagateError(ppm, x, y, green - map_green*0x33, 1);
 			propagateError(ppm, x, y, blue - map_blue*0x33, 2);
		}

	free(ppm->data);
	ppm->data = buf;
	ppm->greyscale = 1;  /* a lie, but it gets the sizes right */
}

/* Propagate error on 4 pixels right-of/below. */
static void propagateError(struct ppm_image *ppm, int x, int y, int error, int plane) {
	int errorSum = 0;
	if (x < ppm->width-1) {
		int shift = 7*error/16;
		PPM_PIXEL(*ppm, x+1, y, plane) = CLAMP(PPM_PIXEL(*ppm, x+1, y, plane) + shift, 0, 255);
		errorSum += shift;
	}
	if (y < ppm->height-1) {
		if (x > 0) {
			int shift = 3*error/16;
			PPM_PIXEL(*ppm, x-1, y+1, plane) = CLAMP(PPM_PIXEL(*ppm, x-1, y+1, plane) + shift, 0, 255);
			errorSum += shift;
		}
		int shift = 5*error/16;
		PPM_PIXEL(*ppm, x, y+1, plane) = CLAMP(PPM_PIXEL(*ppm, x, y+1, plane) + shift, 0, 255);
		errorSum += shift;
		if (x < ppm->width-1) {
			shift = error - errorSum;
			PPM_PIXEL(*ppm, x+1, y+1, plane) = CLAMP(PPM_PIXEL(*ppm, x+1, y+1, plane) + shift, 0, 255);
		}
	}
}
